use crate::{
    drate, mux, opcodes, status, AdControl, Ads1256, BlockingDelay, CalibrationCommand, Conversion,
    DataRate, Gain, Multiplexer, Status,
};
use embedded_hal::digital::OutputPin;
use embedded_hal::spi;
use embedded_hal::spi::SpiBus;
use embedded_hal_async::digital::Wait;
use physical_node::GPIO_ERROR_MSG;
use physical_node::spi::{end_spi, end_spi_if_err};

impl<DelayerT, SST, DrdyT> Ads1256<DelayerT, SST, DrdyT>
where
    DelayerT: BlockingDelay,
    SST: OutputPin,
    DrdyT: Wait,
{
    /// Enter read data continuous mode (wait for drdy low, issue RDATAC command)
    #[inline]
    pub async fn start_rdatac<SpiT: SpiBus>(
        &mut self,
        spi: &mut SpiT,
    ) -> Result<(), <SpiT as spi::ErrorType>::Error> {
        self.data_ready.wait_for_low().await.expect(GPIO_ERROR_MSG);
        self.slave_select.set_low().expect(GPIO_ERROR_MSG);
        let spi_op_result = spi.write(&[opcodes::RDATAC]);
        end_spi(&mut self.slave_select, spi, spi_op_result)?;

        Ok(())
    }

    /// Exit read data continuous mode (wait for drdy low, issue SDATAC command)
    #[inline]
    pub async fn stop_rdatac<SpiT: SpiBus>(
        &mut self,
        spi: &mut SpiT,
    ) -> Result<(), <SpiT as spi::ErrorType>::Error> {
        self.data_ready.wait_for_low().await.expect(GPIO_ERROR_MSG);
        self.slave_select.set_low().expect(GPIO_ERROR_MSG);
        let spi_op_result = spi.write(&[opcodes::SDATAC]);
        end_spi(&mut self.slave_select, spi, spi_op_result)?;

        Ok(())
    }

    pub async fn read_data<SpiT: SpiBus>(
        &mut self,
        spi: &mut SpiT,
    ) -> Result<Conversion, <SpiT as spi::ErrorType>::Error> {
        self.data_ready.wait_for_low().await.expect(GPIO_ERROR_MSG);
        self.slave_select.set_low().expect(GPIO_ERROR_MSG);
        let mut buffer = [0u8; 3];
        let spi_op_result = spi.read(&mut buffer);
        end_spi(&mut self.slave_select, spi, spi_op_result)?;

        Ok(Conversion::from_reading(buffer))
    }

    /// Issues RDATA command and receives result, does not handle any housekeeping such as setting
    /// slave select or flushing the bus on successful completion.
    #[inline]
    pub fn raw_cmd_read_data<SpiT: SpiBus>(
        &mut self,
        spi: &mut SpiT,
    ) -> Result<Conversion, <SpiT as spi::ErrorType>::Error> {
        let spi_op_result = spi.write(&[opcodes::RDATA]);
        end_spi_if_err(&mut self.slave_select, spi, spi_op_result)?;
        self.delayer.t6_delay();
        let mut buffer = [0u8; 3];
        spi.read(&mut buffer)?;

        Ok(Conversion::from_reading(buffer))
    }

    /// Reads the digital conversion value.
    /// Action sequence:
    /// 1. Wait for data_ready to go low
    /// 1. Issue RDATA command
    /// 1. Read the conversion value
    #[inline]
    pub async fn cmd_read_data<SpiT: SpiBus>(
        &mut self,
        spi: &mut SpiT,
    ) -> Result<Conversion, <SpiT as spi::ErrorType>::Error> {
        self.data_ready.wait_for_low().await.expect(GPIO_ERROR_MSG);
        self.slave_select.set_low().expect(GPIO_ERROR_MSG);
        let spi_op_result = self.raw_cmd_read_data(spi);

        end_spi(&mut self.slave_select, spi, spi_op_result)
    }

    #[inline]
    pub fn conversion_init<SpiT: SpiBus>(
        &mut self,
        spi: &mut SpiT,
    ) -> Result<(), <SpiT as spi::ErrorType>::Error> {
        self.slave_select.set_low().expect(GPIO_ERROR_MSG);
        // Issue sync command
        let spi_op_result = spi.write(&[opcodes::SYNC]);
        end_spi_if_err(&mut self.slave_select, spi, spi_op_result)?;
        // Issue wakeup command
        let spi_op_result = spi.write(&[opcodes::WAKEUP]);
        end_spi(&mut self.slave_select, spi, spi_op_result)
    }

    //----- Configure Convert Functions ----------------------------------

    /// Function for configuring and reading AD conversions using auto-calibration. If
    /// only the inputs are changed and no other configuration, auto-calibration will not be
    /// run.
    ///
    /// Action sequence:
    /// 1. Switch inputs and optionally adjust different configuration parameters.
    /// 1. If only the input was switched without configuration changes.
    ///     1. Issue sync command followed by wakeup command
    /// 1. Else, auto-calibration will take place
    /// 1. Wait for data_ready low
    /// 1. RDATA command (read the conversion value)
    /// 1. Optionally enter standby mode
    ///
    /// **WARNING:** Auto-calibration must be enabled for intended functionality when changing
    /// [Status], [AdControl], or [DataRate]. Furthermore if setting [Status] or [AdControl], their
    /// [Buffer] and [Gain] settings must be modified respectively to trigger auto-calibration.
    #[inline]
    pub async fn autocal_convert<SpiT: SpiBus>(
        &mut self,
        spi: &mut SpiT,
        input: Multiplexer,
        status: Option<Status>,
        ad_control: Option<AdControl>,
        data_rate: Option<DataRate>,
        standby: bool,
    ) -> Result<Conversion, <SpiT as spi::ErrorType>::Error> {
        self.slave_select.set_low().expect(GPIO_ERROR_MSG);
        match (status, ad_control, data_rate) {
            // Only modifying the multiplexer, not changing any configuration
            (None, None, None) => {
                self._none_config(spi, input)?;
                self._manual_conversion_init(spi)?;
                self._read_when_rdy(spi, standby).await
            },
            // Modifying status (toggle buffer) and changing multiplexer
            (Some(status), None, None) => {
                self._status_config(spi, input, status)?;
                self._read_when_rdy(spi, standby).await
            },
            // Modifying AD control (gain) and changing multiplexer
            (None, Some(ad_control), None) => {
                self._ad_config(spi, input, ad_control)?;
                self._read_when_rdy(spi, standby).await
            },
            // Modifying data rate and change multiplexer
            (None, None, Some(data_rate)) => {
                self._drate_config(spi, input, data_rate)?;
                self._read_when_rdy(spi, standby).await
            },
            // Modifying status (toggle buffer), AD control (gain), and changing multiplexer
            (Some(status), Some(ad_control), None) => {
                self._status_ad_config(spi, input, status, ad_control)?;
                self._read_when_rdy(spi, standby).await
            },
            // Modifying status (toggle buffer), data rate, and changing multiplexer
            (Some(status), None, Some(data_rate)) => {
                self._status_drate_config(spi, input, status, data_rate)?;
                self._read_when_rdy(spi, standby).await
            },
            // Modifying AD control (gain), data rate, and changing multiplexer
            (None, Some(ad_control), Some(data_rate)) => {
                self._ad_drate_config(spi, input, ad_control, data_rate)?;
                self._read_when_rdy(spi, standby).await
            },
            // Modifying status (toggle buffer), AD control (gain), data rate, and changing
            // multiplexer
            (Some(status), Some(ad_control), Some(data_rate)) => {
                self._all_config(spi, input, status, ad_control, data_rate)?;
                self._read_when_rdy(spi, standby).await
            },
        }
    }

    /// Function for periodically configuring and reading AD conversions using stored calibration
    /// values. If only the inputs are changed and no other configuration, auto-calibration will
    /// not be run. For immediate continuous multiplexing see [Ads1256::loadcal_convert_next].
    ///
    /// Action sequence:
    /// 1. Switch inputs and optionally adjust different configuration parameters.
    /// 1. Issue sync command followed by wakeup command
    /// 1. Wait for data_ready low
    /// 1. RDATA command (read the conversion value)
    /// 1. Optionally enter standby mode
    ///
    /// **WARNING:** Auto-calibration must be disabled for intended functionality when changing
    /// [Status], [AdControl], or [DataRate].
    #[inline]
    pub async fn loadcal_convert<SpiT: SpiBus>(
        &mut self,
        spi: &mut SpiT,
        input: Multiplexer,
        calibration: Option<&CalibrationCommand>,
        status: Option<Status>,
        ad_control: Option<AdControl>,
        data_rate: Option<DataRate>,
        standby: bool,
    ) -> Result<Conversion, <SpiT as spi::ErrorType>::Error> {
        self.slave_select.set_low().expect(GPIO_ERROR_MSG);
        match (status, ad_control, data_rate) {
            // Only modifying the multiplexer, not changing any configuration
            (None, None, None) => {
                self._none_config(spi, input)?;
                self._manual_conversion_init(spi)?;
                self._read_when_rdy(spi, standby).await
            },
            // Modifying status (toggle buffer) and changing multiplexer
            (Some(status), None, None) => {
                self._status_config(spi, input, status)?;
                self._loadcal_init(spi, calibration)?;
                self._read_when_rdy(spi, standby).await
            },
            // Modifying AD control (gain) and changing multiplexer
            (None, Some(ad_control), None) => {
                self._ad_config(spi, input, ad_control)?;
                self._loadcal_init(spi, calibration)?;
                self._read_when_rdy(spi, standby).await
            },
            // Modifying data rate and change multiplexer
            (None, None, Some(data_rate)) => {
                self._drate_config(spi, input, data_rate)?;
                self._loadcal_init(spi, calibration)?;
                self._read_when_rdy(spi, standby).await
            },
            // Modifying status (toggle buffer), AD control (gain), and changing multiplexer
            (Some(status), Some(ad_control), None) => {
                self._status_ad_config(spi, input, status, ad_control)?;
                self._loadcal_init(spi, calibration)?;
                self._read_when_rdy(spi, standby).await
            },
            // Modifying status (toggle buffer), data rate, and changing multiplexer
            (Some(status), None, Some(data_rate)) => {
                self._status_drate_config(spi, input, status, data_rate)?;
                self._loadcal_init(spi, calibration)?;
                self._read_when_rdy(spi, standby).await
            },
            // Modifying AD control (gain), data rate, and changing multiplexer
            (None, Some(ad_control), Some(data_rate)) => {
                self._ad_drate_config(spi, input, ad_control, data_rate)?;
                self._loadcal_init(spi, calibration)?;
                self._read_when_rdy(spi, standby).await
            },
            // Modifying status (toggle buffer), AD control (gain), data rate, and changing
            // multiplexer
            (Some(status), Some(ad_control), Some(data_rate)) => {
                self._all_config(spi, input, status, ad_control, data_rate)?;
                self._loadcal_init(spi, calibration)?;
                self._read_when_rdy(spi, standby).await
            },
        }
    }

    /// Function for rapidly configuring and reading AD conversions using stored calibration
    /// values. If only the inputs are changed and no other configuration, auto-calibration will
    /// not be run. For periodic reading see [Ads1256::loadcal_convert].
    ///
    /// Action sequence:
    /// 1. Wait for data_ready low
    /// 1. RDATA command (read the conversion value)
    /// 1. Switch inputs and optionally adjust different configuration parameters.
    /// 1. Issue sync command followed by wakeup command
    ///
    /// **WARNING:** Auto-calibration must be disabled for intended functionality when changing
    /// [Status], [AdControl], or [DataRate].
    #[inline]
    pub async fn loadcal_convert_next<SpiT: SpiBus>(
        &mut self,
        spi: &mut SpiT,
        next_input: Multiplexer,
        next_calibration: Option<&CalibrationCommand>,
        next_status: Option<Status>,
        next_ad_control: Option<AdControl>,
        next_data_rate: Option<DataRate>,
        standby: bool,
    ) -> Result<Conversion, <SpiT as spi::ErrorType>::Error> {
        self.data_ready.wait_for_low().await.expect(GPIO_ERROR_MSG);
        self._loadcal_convert_next(
            spi,
            next_input,
            next_calibration,
            next_status,
            next_ad_control,
            next_data_rate,
            standby,
        )
        .await
    }

    #[inline(always)]
    pub(crate) async fn _loadcal_convert_next<SpiT: SpiBus>(
        &mut self,
        spi: &mut SpiT,
        next_input: Multiplexer,
        next_calibration: Option<&CalibrationCommand>,
        next_status: Option<Status>,
        next_ad_control: Option<AdControl>,
        next_data_rate: Option<DataRate>,
        standby: bool,
    ) -> Result<Conversion, <SpiT as spi::ErrorType>::Error> {
        self.slave_select.set_low().expect(GPIO_ERROR_MSG);
        match (next_status, next_ad_control, next_data_rate) {
            // Only modifying the multiplexer, not changing any configuration
            (None, None, None) => {
                self._none_config(spi, next_input)?;
                self._manual_conversion_init(spi)?;
                self._read_mux_conversion(spi, standby)
            },
            // Modifying status (toggle buffer) and changing multiplexer
            (Some(status), None, None) => {
                self._status_config(spi, next_input, status)?;
                self._loadcal_init(spi, next_calibration)?;
                self._read_mux_conversion(spi, standby)
            },
            // Modifying AD control (gain) and changing multiplexer
            (None, Some(ad_control), None) => {
                self._ad_config(spi, next_input, ad_control)?;
                self._loadcal_init(spi, next_calibration)?;
                self._read_mux_conversion(spi, standby)
            },
            // Modifying data rate and change multiplexer
            (None, None, Some(data_rate)) => {
                self._drate_config(spi, next_input, data_rate)?;
                self._loadcal_init(spi, next_calibration)?;
                self._read_mux_conversion(spi, standby)
            },
            // Modifying status (toggle buffer), AD control (gain), and changing multiplexer
            (Some(status), Some(ad_control), None) => {
                self._status_ad_config(spi, next_input, status, ad_control)?;
                self._loadcal_init(spi, next_calibration)?;
                self._read_mux_conversion(spi, standby)
            },
            // Modifying status (toggle buffer), data rate, and changing multiplexer
            (Some(status), None, Some(data_rate)) => {
                self._status_drate_config(spi, next_input, status, data_rate)?;
                self._loadcal_init(spi, next_calibration)?;
                self._read_mux_conversion(spi, standby)
            },
            // Modifying AD control (gain), data rate, and changing multiplexer
            (None, Some(ad_control), Some(data_rate)) => {
                self._ad_drate_config(spi, next_input, ad_control, data_rate)?;
                self._loadcal_init(spi, next_calibration)?;
                self._read_mux_conversion(spi, standby)
            },
            // Modifying status (toggle buffer), AD control (gain), data rate, and changing
            // multiplexer
            (Some(status), Some(ad_control), Some(data_rate)) => {
                self._all_config(spi, next_input, status, ad_control, data_rate)?;
                self._loadcal_init(spi, next_calibration)?;
                self._read_mux_conversion(spi, standby)
            },
        }
    }

    /// Read a pre-defined number of samples. This will use read continuous command, the ADS1256 must be
    /// properly configure before calling this function. This uses [BlockingDelay::t11_1_delay] before running so as it's
    /// almost always used after configuring the ADS1256 for a specific channel.
    pub async fn multi_sample<SpiT: SpiBus, const NUM_SAMPLES: usize>(
        &mut self,
        spi: &mut SpiT,
    ) -> Result<[Conversion; NUM_SAMPLES], <SpiT as spi::ErrorType>::Error> {
        let mut samples: [Conversion; NUM_SAMPLES] = [0.into(); NUM_SAMPLES];
        self.delayer.t11_1_delay();

        self.start_rdatac(spi).await?;
        for i in 0..NUM_SAMPLES {
            samples[i] = self.read_data(spi).await?;
        }
        self.delayer.t6_delay();
        self.stop_rdatac(spi).await?;

        Ok(samples)
    }

    #[inline]
    pub(crate) fn _none_config<SpiT: SpiBus>(
        &mut self,
        spi: &mut SpiT,
        input: Multiplexer,
    ) -> Result<(), <SpiT as spi::ErrorType>::Error> {
        // Change inputs
        let buffer = [0u8, 0u8, input.0];
        let spi_op_result = self.raw_write_registers(spi, mux::ADDRESS, buffer);
        end_spi_if_err(&mut self.slave_select, spi, spi_op_result)
    }

    #[inline]
    pub(crate) fn _loadcal_init<SpiT: SpiBus>(
        &mut self,
        spi: &mut SpiT,
        calibration: Option<&CalibrationCommand>,
    ) -> Result<(), <SpiT as spi::ErrorType>::Error> {
        self.delayer.t11_1_delay();
        if let Some(calibration) = calibration {
            let spi_op_result = spi.write(calibration.into());
            end_spi_if_err(&mut self.slave_select, spi, spi_op_result)?;
        }
        self._manual_conversion_init(spi)
    }

    #[inline]
    pub(crate) fn _manual_conversion_init<SpiT: SpiBus>(
        &mut self,
        spi: &mut SpiT,
    ) -> Result<(), <SpiT as spi::ErrorType>::Error> {
        self.delayer.t11_1_delay();
        // Since we did not change configuration, ADS1256 will not auto-calibrate so we
        // need to sync, wakeup
        // Issue sync command
        let spi_op_result = spi.write(&[opcodes::SYNC]);
        end_spi_if_err(&mut self.slave_select, spi, spi_op_result)?;
        // Issue wakeup command
        let spi_op_result = spi.write(&[opcodes::WAKEUP]);
        end_spi_if_err(&mut self.slave_select, spi, spi_op_result)
    }

    #[inline]
    pub(crate) fn _status_config<SpiT: SpiBus>(
        &mut self,
        spi: &mut SpiT,
        input: Multiplexer,
        status: Status,
    ) -> Result<(), <SpiT as spi::ErrorType>::Error> {
        let buffer = [0u8, 0u8, status.0, input.0];
        let spi_op_result = self.raw_write_registers(spi, status::ADDRESS, buffer);

        end_spi_if_err(&mut self.slave_select, spi, spi_op_result)
    }

    #[inline]
    pub(crate) fn _ad_config<SpiT: SpiBus>(
        &mut self,
        spi: &mut SpiT,
        input: Multiplexer,
        ad_control: AdControl,
    ) -> Result<(), <SpiT as spi::ErrorType>::Error> {
        let buffer = [0u8, 0u8, input.0, ad_control.0];
        let spi_op_result = self.raw_write_registers(spi, mux::ADDRESS, buffer);

        end_spi_if_err(&mut self.slave_select, spi, spi_op_result)
    }

    #[inline]
    pub(crate) fn _drate_config<SpiT: SpiBus>(
        &mut self,
        spi: &mut SpiT,
        input: Multiplexer,
        data_rate: DataRate,
    ) -> Result<(), <SpiT as spi::ErrorType>::Error> {
        // Switch inputs
        let mut buffer = [0u8, 0u8, input.0];
        let spi_op_result = self.raw_write_registers(spi, mux::ADDRESS, buffer);
        end_spi_if_err(&mut self.slave_select, spi, spi_op_result)?;
        self.delayer.t11_1_delay();
        // Modify data rate
        buffer[2] = data_rate as u8;
        let spi_op_result = self.raw_write_registers(spi, drate::ADDRESS, buffer);

        end_spi_if_err(&mut self.slave_select, spi, spi_op_result)
    }

    #[inline]
    pub(crate) fn _status_ad_config<SpiT: SpiBus>(
        &mut self,
        spi: &mut SpiT,
        input: Multiplexer,
        status: Status,
        ad_control: AdControl,
    ) -> Result<(), <SpiT as spi::ErrorType>::Error> {
        let buffer = [0u8, 0u8, status.0, input.0, ad_control.0];
        let spi_op_result = self.raw_write_registers(spi, status::ADDRESS, buffer);

        end_spi_if_err(&mut self.slave_select, spi, spi_op_result)
    }

    #[inline]
    pub(crate) fn _status_drate_config<SpiT: SpiBus>(
        &mut self,
        spi: &mut SpiT,
        input: Multiplexer,
        status: Status,
        data_rate: DataRate,
    ) -> Result<(), <SpiT as spi::ErrorType>::Error> {
        // Modify status and switch inputs
        let buffer = [0u8, 0u8, status.0, input.0];
        let spi_op_result = self.raw_write_registers(spi, status::ADDRESS, buffer);
        end_spi_if_err(&mut self.slave_select, spi, spi_op_result)?;
        self.delayer.t11_1_delay();
        // Modify data rate
        let buffer = [0u8, 0u8, data_rate as u8];
        let spi_op_result = self.raw_write_registers(spi, drate::ADDRESS, buffer);

        end_spi_if_err(&mut self.slave_select, spi, spi_op_result)
    }

    #[inline]
    pub(crate) fn _ad_drate_config<SpiT: SpiBus>(
        &mut self,
        spi: &mut SpiT,
        input: Multiplexer,
        ad_control: AdControl,
        data_rate: DataRate,
    ) -> Result<(), <SpiT as spi::ErrorType>::Error> {
        let buffer = [0u8, 0u8, input.0, ad_control.0, data_rate as u8];
        let spi_op_result = self.raw_write_registers(spi, mux::ADDRESS, buffer);

        end_spi_if_err(&mut self.slave_select, spi, spi_op_result)
    }

    #[inline]
    pub(crate) fn _all_config<SpiT: SpiBus>(
        &mut self,
        spi: &mut SpiT,
        input: Multiplexer,
        status: Status,
        ad_control: AdControl,
        data_rate: DataRate,
    ) -> Result<(), <SpiT as spi::ErrorType>::Error> {
        let buffer = [0u8, 0u8, status.0, input.0, ad_control.0, data_rate as u8];
        let spi_op_result = self.raw_write_registers(spi, status::ADDRESS, buffer);

        end_spi_if_err(&mut self.slave_select, spi, spi_op_result)
    }

    #[inline]
    async fn _read_when_rdy<SpiT: SpiBus>(
        &mut self,
        spi: &mut SpiT,
        standby: bool,
    ) -> Result<Conversion, <SpiT as spi::ErrorType>::Error> {
        // Wait for data ready low
        self.data_ready.wait_for_low().await.expect(GPIO_ERROR_MSG);
        // Read data
        self._read_mux_conversion(spi, standby)
    }

    #[inline]
    pub(crate) fn _read_mux_conversion<SpiT: SpiBus>(
        &mut self,
        spi: &mut SpiT,
        standby: bool,
    ) -> Result<Conversion, <SpiT as spi::ErrorType>::Error> {
        // Read data
        let spi_op_result = self.raw_cmd_read_data(spi);
        let cvalue = end_spi_if_err(&mut self.slave_select, spi, spi_op_result)?;
        // if standby is true enter standby mode
        if standby {
            let spi_op_result = spi.write(&[opcodes::STANDBY]);
            end_spi_if_err(&mut self.slave_select, spi, spi_op_result)?;
        }
        // end SPI
        self.slave_select.set_high().expect(GPIO_ERROR_MSG);
        spi.flush()?;

        Ok(cvalue)
    }
}
