pub const REFERENCE_VOLTS: f32 = 2.5;
pub const MAX_CONVERSION_VALUE: i32 = 8_388_607;

pub mod defaults {
    //----- Public ----------------------------------
    pub const SPI_CLK_HZ: u32 = 1_920_000;
}

pub mod opcodes {
    /// Completes SYNC and Exits Standby Mode.
    pub const WAKEUP: u8 = 0b00000000;
    /// Read Data.
    pub const RDATA: u8 = 0b00000001;
    /// Read Data Continuously.
    pub const RDATAC: u8 = 0b00000011;
    /// Stop Read Data Continuously
    pub const SDATAC: u8 = 0b00001111;
    /// Read from register at rrrr. The command is only the first 4 bits, the last 4 bits need to be
    /// changed to the register address.
    pub const RREG: u8 = 0b0001_0000;
    /// Write to register at rrrr. The command is only the first 4 bits, the last 4 bits need to be
    /// changed to the register address.
    pub const WREG: u8 = 0b0101_0000;
    /// Offset and Gain Self-Calibration.
    pub const SELFCAL: u8 = 0b11110000;
    /// Offset Self-Calibration.
    pub const SELFOCAL: u8 = 0b11110001;
    /// Gain Self-Calibration.
    pub const SELFGCAL: u8 = 0b11110010;
    /// System Offset Calibration.
    pub const SYSOCAL: u8 = 0b11110011;
    /// System Gain Calibration.
    pub const SYSGCAL: u8 = 0b11110100;
    /// Synchronize the A/D Conversion.
    pub const SYNC: u8 = 0b11111100;
    /// Begin Standby Mode.
    pub const STANDBY: u8 = 0b11111101;
    /// Reset to Power-Up Values.
    pub const RESET: u8 = 0b11111110;
}

/// Status register.
pub mod status {
    /// Address of the STATUS register.
    pub const ADDRESS: u8 = 0b0000_0000;

    #[cfg_attr(feature = "defmt", derive(defmt::Format))]
    #[derive(Copy, Clone, Eq, PartialEq, Debug)]
    #[repr(u8)]
    pub enum Buffer {
        Enabled  = 0b000000_1_0,
        Disabled = 0,
    }

    impl Buffer {
        pub const MASK: u8 = 0b000000_1_0;
    }

    #[cfg_attr(feature = "defmt", derive(defmt::Format))]
    #[derive(Copy, Clone, Eq, PartialEq, Debug)]
    #[repr(u8)]
    pub enum AutoCal {
        Enabled  = 0b00000_1_00,
        Disabled = 0,
    }

    impl AutoCal {
        pub const MASK: u8 = 0b00000_1_00;
    }

    /// Input data is always shifted in most significant byte and bit first. Output data is always
    /// shifted out most significant byte first. The [BitOrder] only controls the bit order of the
    /// output data within the byte.
    #[cfg_attr(feature = "defmt", derive(defmt::Format))]
    #[derive(Copy, Clone, Eq, PartialEq, Debug)]
    #[repr(u8)]
    pub enum BitOrder {
        /// Most significant bit first.
        MostSigFirst  = 0,
        /// Least significant bit first.
        LeastSigFirst = 0b0000_1_000,
    }

    impl BitOrder {
        pub const MASK: u8 = 0b0000_1_000;
    }
}

/// Input multiplexer control register.
pub mod mux {
    /// address of the MUX register.
    pub const ADDRESS: u8 = 0b0000_0001;

    #[cfg_attr(feature = "defmt", derive(defmt::Format))]
    #[derive(Copy, Clone, Eq, PartialEq, Debug)]
    #[repr(u8)]
    pub enum MuxInput {
        AIn0   = 0b0000_0000,
        AIn1   = 0b0000_0001,
        AIn2   = 0b0000_0010,
        AIn3   = 0b0000_0011,
        AIn4   = 0b0000_0100,
        AIn5   = 0b0000_0101,
        AIn6   = 0b0000_0110,
        AIn7   = 0b0000_0111,
        Common = 0b0000_1000,
    }

    impl MuxInput {
        pub const NEGATIVE_MASK: u8 = 0b0000_1111;
        pub const POSITIVE_MASK: u8 = 0b1111_0000;
    }
}

/// A/D control register.
pub mod adcon {
    /// Address of the ADCON register.
    pub const ADDRESS: u8 = 0b0000_0010;

    //TODO: Fitting the value and encoding in 16 bits can probably be done more cleanly if enum
    // variants become types in a later version of Rust.
    /// **Warning:** casting to an integer will likely not yield expected result. Use `value()` or
    /// `encoding()` methods instead.
    #[cfg_attr(feature = "defmt", derive(defmt::Format))]
    #[derive(Copy, Clone, Eq, PartialEq, Debug)]
    #[repr(u8)]
    pub enum Gain {
        X1  = 0b00000_000,
        X2  = 0b00000_001,
        X4  = 0b00000_010,
        X8  = 0b00000_011,
        X16 = 0b00000_100,
        X32 = 0b00000_101,
        X64 = 0b00000_110,
    }

    impl Gain {
        /// Value to use binary & operator on the full ADCON byte to isolate the gain.
        pub const MASK: u8 = 0b00000_111;
        pub const ALT_X64: u8 = 0b00000_111;

        /// The integer value of this [Gain].
        ///
        /// ```
        /// use ads1256_types::Gain;
        ///
        /// assert_eq!(Gain::X1.value(), 1);
        /// assert_eq!(Gain::X2.value(), 2);
        /// assert_eq!(Gain::X4.value(), 4);
        /// assert_eq!(Gain::X8.value(), 8);
        /// assert_eq!(Gain::X16.value(), 16);
        /// assert_eq!(Gain::X32.value(), 32);
        /// assert_eq!(Gain::X64.value(), 64);
        /// ```
        #[inline(always)]
        pub const fn value(self) -> u8 {
            // Multiply 1 by 2 the number of times specified in the encoding (e.g. for gain of 32,
            // (1 * 2) 5 times (5 is the encoding for 32) is 32).
            1 << (self as u8)
        }
    }

    /// Sensor detect current sources.
    /// For testing that sensor is still connected and able to pass current.
    #[cfg_attr(feature = "defmt", derive(defmt::Format))]
    #[derive(Copy, Clone, Eq, PartialEq, Debug)]
    #[repr(u8)]
    pub enum Sdcs {
        Off = 0b000_00_000,
        /// 0.5µA
        C05 = 0b000_01_000,
        /// 2µA
        C2  = 0b000_10_000,
        /// 10µA
        C10 = 0b000_11_000,
    }

    impl Sdcs {
        /// Value to use binary & operator on the full ADCON byte to isolate the SDCS.
        pub const MASK: u8 = 0b000_11_000;
    }

    /// Ads1256 master clock cycle frequency outputted through GPIO.
    #[allow(non_camel_case_types)]
    #[cfg_attr(feature = "defmt", derive(defmt::Format))]
    #[derive(Copy, Clone, Eq, PartialEq, Debug)]
    #[repr(u8)]
    pub enum ClockOut {
        Off     = 0b0_00_00000,
        /// Equal to clock in frequency.
        ClkIn   = 0b0_01_00000,
        /// Equal to clock in frequency / 2.
        ClkIn_2 = 0b0_10_00000,
        /// Equal to clock in frequency / 4.
        ClkIn_4 = 0b0_11_00000,
    }

    impl ClockOut {
        /// Value to use binary & operator on the full ADCON byte to isolate the clock out
        /// frequency.
        pub const MASK: u8 = 0b0_11_00000;
    }
}

/// A/D data rate register.
pub mod drate {
    /// Address of the DRATE register.
    pub const ADDRESS: u8 = 0b0000_0011;

    #[cfg_attr(feature = "defmt", derive(defmt::Format))]
    #[derive(Copy, Clone, Eq, PartialEq, Debug)]
    #[repr(u8)]
    pub enum DataRate {
        Sps2_5   = 0b00000011,
        Sps5     = 0b00010011,
        Sps10    = 0b00100011,
        Sps15    = 0b00110011,
        Sps25    = 0b01000011,
        Sps30    = 0b01010011,
        Sps50    = 0b01100011,
        Sps60    = 0b01110010,
        Sps100   = 0b10000010,
        Sps500   = 0b10010010,
        Sps1000  = 0b10100001,
        Sps2000  = 0b10110000,
        Sps3750  = 0b11000000,
        Sps7500  = 0b11010000,
        Sps15000 = 0b11100000,
        Sps30000 = 0b11110000,
    }
}

pub mod gpio {
    /// Address of the IO register.
    pub const ADDRESS: u8 = 0b0000_0100;

    #[cfg_attr(feature = "defmt", derive(defmt::Format))]
    #[derive(Copy, Clone, Eq, PartialEq, Debug)]
    #[repr(u8)]
    pub enum DioDirection {
        Output = 0,
        Input  = 1,
    }

    impl DioDirection {
        pub const ALL_MASK: u8 = 0b1111_0000;
        pub const IO0_MASK: u8 = 0b10000;
        pub const IO1_MASK: u8 = 0b100000;
        pub const IO2_MASK: u8 = 0b1000000;
        pub const IO3_MASK: u8 = 0b10000000;
    }

    #[cfg_attr(feature = "defmt", derive(defmt::Format))]
    #[derive(Copy, Clone, Eq, PartialEq, Debug)]
    #[repr(u8)]
    pub enum DState {
        Low  = 0,
        High = 1,
    }

    impl DState {
        pub const ALL_MASK: u8 = 0b0000_1111;
        pub const IO0_MASK: u8 = 0b1;
        pub const IO1_MASK: u8 = 0b10;
        pub const IO2_MASK: u8 = 0b100;
        pub const IO3_MASK: u8 = 0b1000;
    }
}

/// Offset calibration byte 0.
pub mod ofc0 {
    /// Address of the OFC0 register.
    pub const ADDRESS: u8 = 0b0000_0101;
}

/// Offset calibration byte 1.
pub mod ofc1 {
    /// Address of the OFC1 register.
    pub const ADDRESS: u8 = 0b0000_0110;
}

/// Offset calibration byte 2.
pub mod ofc2 {
    /// Address of the OFC2 register.
    pub const ADDRESS: u8 = 0b0000_0111;
}

/// Full scale calibration byte 0
pub mod fsc0 {
    /// Address of the FSC0 register.
    pub const ADDRESS: u8 = 0b0000_1000;
}

/// Full scale calibration byte 0
pub mod fsc1 {
    /// Address of the FSC1 register.
    pub const ADDRESS: u8 = 0b0000_1001;
}

/// Full scale calibration byte 0
pub mod fsc2 {
    /// Address of the FSC2 register.
    pub const ADDRESS: u8 = 0b0000_1010;
}
