use crate::{
    ofc0, opcodes, AutoCal, BitOrder, Buffer, ClockOut, DState, DataRate, DioDirection, Gain,
    MuxInput, Sdcs,
};
use core::mem;

// ---------------------------------------------------------------------------------------------------------------------
// ----- Calibration ------------------------
// ---------------------------------------------------------------------------------------------------------------------
#[cfg_attr(feature = "defmt", derive(defmt::Format))]
#[derive(Copy, Clone, Eq, PartialEq, Debug)]
#[repr(transparent)]
pub struct OffsetCalibration(pub [u8; 3]);

#[cfg_attr(feature = "defmt", derive(defmt::Format))]
#[derive(Copy, Clone, Eq, PartialEq, Debug)]
#[repr(transparent)]
pub struct GainCalibration(pub [u8; 3]);

#[cfg_attr(feature = "defmt", derive(defmt::Format))]
#[derive(Copy, Clone, Eq, PartialEq, Debug)]
#[repr(C)]
pub struct AllCalibration {
    pub offset: OffsetCalibration,
    pub gain: GainCalibration,
}

impl From<[u8; 6]> for AllCalibration {
    /// This array must be formatted as offset cal bytes [0..2] followed by gain cal bytes [0..2]
    /// for resulting [AllCalibration] to be meaningful.
    #[inline(always)]
    fn from(array: [u8; 6]) -> Self {
        unsafe { mem::transmute::<[u8; 6], Self>(array) }
    }
}

impl Into<[u8; 6]> for AllCalibration {
    #[inline(always)]
    fn into(self) -> [u8; 6] {
        unsafe { mem::transmute::<Self, [u8; 6]>(self) }
    }
}

/// The complete formatted command to write all calibration values to ADS1256 registers.
/// Prepending the write command bytes to the calibration value is a fairly costly operation and in
/// practice the only real use of calibration values is to store them and write them to the ADS1256.
/// When multiplexing inputs we store calibration infrequently and write it frequently so it's
/// better to take the overhead of prepending the command bytes when we store, not when we write.
#[derive(Copy, Clone, Eq, PartialEq)]
#[repr(C)]
pub struct CalibrationCommand {
    command: [u8; 2],
    calibration: AllCalibration,
}

impl From<AllCalibration> for CalibrationCommand {
    fn from(calibration: AllCalibration) -> Self {
        Self {
            command: [opcodes::WREG | ofc0::ADDRESS, 5],
            calibration,
        }
    }
}

impl Into<[u8; 8]> for CalibrationCommand {
    #[inline(always)]
    fn into(self) -> [u8; 8] {
        unsafe { mem::transmute::<Self, [u8; 8]>(self) }
    }
}

impl<'a> Into<&'a [u8]> for &'a CalibrationCommand {
    #[inline(always)]
    fn into(self) -> &'a [u8] {
        unsafe { mem::transmute::<Self, &'a [u8; 8]>(self) }
    }
}

// ---------------------------------------------------------------------------------------------------------------------
// ----- Config ------------------------
// ---------------------------------------------------------------------------------------------------------------------
#[derive(Copy, Clone, Eq, PartialEq)]
#[repr(C)]
pub struct Config {
    pub status: Status,
    pub multiplexer: Multiplexer,
    pub ad_control: AdControl,
    pub data_rate: DataRate,
    pub digital_io: DigitalIo,
}

impl Config {
    // We want this internal and are only using it in [Ads1256::read_config].
    #[inline]
    pub const fn from_bytes(bytes: [u8; 5]) -> Self {
        Self {
            status: Status(bytes[0]),
            multiplexer: Multiplexer(bytes[1]),
            ad_control: AdControl(bytes[2]),
            data_rate: DataRate::from_byte(bytes[3]),
            digital_io: DigitalIo(bytes[4]),
        }
    }
}

#[cfg(feature = "defmt")]
impl defmt::Format for Config {
    #[rustfmt::skip]
    fn format(&self, fmt: defmt::Formatter) {
        defmt::write!(
            fmt,
            "Config {{\n\
                \t{},\n\
                \t{},\n\
                \t{},\n\
                \tdate_rate: {},\n\
                \t{}\n\
            }}",
            self.status,
            self.multiplexer,
            self.ad_control,
            self.data_rate,
            self.digital_io
        )
    }
}

// ---------------------------------------------------------------------------------------------------------------------
// ----- Status ------------------------
// ---------------------------------------------------------------------------------------------------------------------
#[derive(Copy, Clone, Eq, PartialEq)]
#[repr(transparent)]
pub struct Status(pub u8);

impl Status {
    /// Creates a new [Status] with all writable properties explicitly set.
    #[inline(always)]
    pub const fn setting(buffer: Buffer, auto_calibration: AutoCal, bit_order: BitOrder) -> Status {
        Status(buffer as u8 | auto_calibration as u8 | bit_order as u8)
    }

    /// returns a copy of the [Status] with the buffer setting replaced by the [setting] parameter.
    #[inline(always)]
    pub const fn with_buffer(self, setting: Buffer) -> Status {
        let zeroed_setting = self.0 & !Buffer::MASK;
        Status(zeroed_setting | setting as u8)
    }

    /// returns a copy of the [Status] with the auto-calibration setting replaced by the [setting]
    /// parameter.
    #[inline(always)]
    pub const fn with_auto_cal(self, setting: AutoCal) -> Status {
        let zeroed_setting = self.0 & !AutoCal::MASK;
        Status(zeroed_setting | setting as u8)
    }

    #[inline(always)]
    pub const fn data_ready(self) -> bool {
        const MASK: u8 = 0b1;
        unsafe { mem::transmute::<u8, bool>(self.0 & MASK) }
    }

    #[inline(always)]
    pub const fn buffer(self) -> Buffer {
        unsafe { mem::transmute::<u8, Buffer>(self.0 & Buffer::MASK) }
    }

    #[inline(always)]
    pub const fn auto_calibration(self) -> AutoCal {
        unsafe { mem::transmute::<u8, AutoCal>(self.0 & AutoCal::MASK) }
    }

    /// Data output bit order.
    ///
    /// Input data is always shifted in most significant byte and bit first. Output data is always
    /// shifted out most significant byte first. The ORDER bit only controls the bit order of the
    /// output data within the byte.
    #[inline(always)]
    pub const fn data_output_bit_order(self) -> BitOrder {
        unsafe { mem::transmute::<u8, BitOrder>(self.0 & BitOrder::MASK) }
    }

    #[inline(always)]
    pub const fn id(self) -> u8 {
        self.0 >> 4
    }
}

impl Into<u8> for Status {
    #[inline(always)]
    fn into(self) -> u8 {
        self.0
    }
}

#[cfg(feature = "defmt")]
impl defmt::Format for Status {
    fn format(&self, fmt: defmt::Formatter) {
        defmt::write!(
            fmt,
            "Status(data_ready: {}, buffer: {}, auto_calibration: {}, data_output_bit_order: {}, \
             id: {})",
            self.data_ready(),
            self.buffer(),
            self.auto_calibration(),
            self.data_output_bit_order(),
            self.id()
        )
    }
}

impl Buffer {
    #[inline(always)]
    pub const fn is_enabled(self) -> bool {
        match self {
            Buffer::Enabled => true,
            Buffer::Disabled => false,
        }
    }
}

impl AutoCal {
    #[inline(always)]
    pub const fn is_enabled(self) -> bool {
        match self {
            AutoCal::Enabled => true,
            AutoCal::Disabled => false,
        }
    }
}
// ---------------------------------------------------------------------------------------------------------------------
// ----- Input Multiplexer ------------------------
// ---------------------------------------------------------------------------------------------------------------------
#[derive(Copy, Clone, Eq, PartialEq)]
#[repr(transparent)]
pub struct Multiplexer(pub u8);

impl Multiplexer {
    #[inline(always)]
    pub const fn setting(positive_input: MuxInput, negative_input: MuxInput) -> Multiplexer {
        Multiplexer(((positive_input as u8) << 4) | negative_input as u8)
    }

    #[inline(always)]
    pub const fn with_positive(self, input: MuxInput) -> Multiplexer {
        let zeroed_positive = self.0 & !MuxInput::POSITIVE_MASK;
        Multiplexer(zeroed_positive | ((input as u8) << 4))
    }

    #[inline(always)]
    pub const fn with_negative(self, input: MuxInput) -> Multiplexer {
        let zeroed_negative = self.0 & !MuxInput::NEGATIVE_MASK;
        Multiplexer(zeroed_negative | input as u8)
    }

    #[inline]
    pub const fn positive_input(self) -> MuxInput {
        const A_IN0: u8 = (MuxInput::AIn0 as u8) << 4;
        const A_IN1: u8 = (MuxInput::AIn1 as u8) << 4;
        const A_IN2: u8 = (MuxInput::AIn2 as u8) << 4;
        const A_IN3: u8 = (MuxInput::AIn3 as u8) << 4;
        const A_IN4: u8 = (MuxInput::AIn4 as u8) << 4;
        const A_IN5: u8 = (MuxInput::AIn5 as u8) << 4;
        const A_IN6: u8 = (MuxInput::AIn6 as u8) << 4;
        const A_IN7: u8 = (MuxInput::AIn7 as u8) << 4;

        match self.0 & MuxInput::POSITIVE_MASK {
            A_IN0 => MuxInput::AIn0,
            A_IN1 => MuxInput::AIn1,
            A_IN2 => MuxInput::AIn2,
            A_IN3 => MuxInput::AIn3,
            A_IN4 => MuxInput::AIn4,
            A_IN5 => MuxInput::AIn5,
            A_IN6 => MuxInput::AIn6,
            A_IN7 => MuxInput::AIn7,
            _ => MuxInput::Common,
        }
    }

    #[inline]
    pub const fn negative_input(self) -> MuxInput {
        const A_IN0: u8 = MuxInput::AIn0 as u8;
        const A_IN1: u8 = MuxInput::AIn1 as u8;
        const A_IN2: u8 = MuxInput::AIn2 as u8;
        const A_IN3: u8 = MuxInput::AIn3 as u8;
        const A_IN4: u8 = MuxInput::AIn4 as u8;
        const A_IN5: u8 = MuxInput::AIn5 as u8;
        const A_IN6: u8 = MuxInput::AIn6 as u8;
        const A_IN7: u8 = MuxInput::AIn7 as u8;

        match self.0 & MuxInput::NEGATIVE_MASK {
            A_IN0 => MuxInput::AIn0,
            A_IN1 => MuxInput::AIn1,
            A_IN2 => MuxInput::AIn2,
            A_IN3 => MuxInput::AIn3,
            A_IN4 => MuxInput::AIn4,
            A_IN5 => MuxInput::AIn5,
            A_IN6 => MuxInput::AIn6,
            A_IN7 => MuxInput::AIn7,
            _ => MuxInput::Common,
        }
    }
}

impl Into<u8> for Multiplexer {
    #[inline(always)]
    fn into(self) -> u8 {
        self.0
    }
}

#[cfg(feature = "defmt")]
impl defmt::Format for Multiplexer {
    fn format(&self, fmt: defmt::Formatter) {
        defmt::write!(
            fmt,
            "Multiplexer(positive_input: {}, negative_input: {})",
            self.positive_input(),
            self.negative_input()
        )
    }
}

// ---------------------------------------------------------------------------------------------------------------------
// ----- A/D Control ------------------------
// ---------------------------------------------------------------------------------------------------------------------
#[derive(Copy, Clone, Eq, PartialEq)]
#[repr(transparent)]
pub struct AdControl(pub u8);

impl AdControl {
    #[inline(always)]
    pub const fn setting(
        gain: Gain,
        sensor_detect_current: Sdcs,
        clock_out: ClockOut,
    ) -> AdControl {
        AdControl(gain as u8 | sensor_detect_current as u8 | clock_out as u8)
    }

    #[inline(always)]
    pub const fn with_gain(self, setting: Gain) -> AdControl {
        let zeroed_setting = self.0 & !Gain::MASK;
        AdControl(zeroed_setting | setting as u8)
    }

    #[inline(always)]
    pub const fn with_sensor_detect_current(self, setting: Sdcs) -> AdControl {
        let zeroed_setting = self.0 & !Sdcs::MASK;
        AdControl(zeroed_setting | setting as u8)
    }

    #[inline(always)]
    pub const fn with_clock_out(self, setting: ClockOut) -> AdControl {
        let zeroed_setting = self.0 & !ClockOut::MASK;
        AdControl(zeroed_setting | setting as u8)
    }

    /// Programmable gain amplifier setting.
    /// *Note: On the ADS1256 0b110 and 0b111 both represent a gain amplification setting of X64 but
    /// in this driver it will always be stored as 0b110.*
    #[inline]
    pub const fn gain(self) -> Gain {
        let masked = self.0 & Gain::MASK;
        if masked == Gain::ALT_X64 {
            Gain::X64
        } else {
            unsafe { mem::transmute::<u8, Gain>(masked) }
        }
    }

    /// Sensor detect current sources. For testing that sensor is still connected and able to pass
    /// current.
    #[inline(always)]
    pub const fn sensor_detect_current(self) -> Sdcs {
        unsafe { mem::transmute::<u8, Sdcs>(self.0 & Sdcs::MASK) }
    }

    /// Ads1256 master clock cycle frequency outputted through GPIO.
    #[inline(always)]
    pub const fn clock_out(self) -> ClockOut {
        unsafe { mem::transmute::<u8, ClockOut>(self.0 & ClockOut::MASK) }
    }
}

impl Into<u8> for AdControl {
    #[inline(always)]
    fn into(self) -> u8 {
        self.0
    }
}

#[cfg(feature = "defmt")]
impl defmt::Format for AdControl {
    fn format(&self, fmt: defmt::Formatter) {
        defmt::write!(
            fmt,
            "AdControl(gain: {}, sensor_detect_current: {}, clock_out: {})",
            self.gain(),
            self.sensor_detect_current(),
            self.clock_out()
        )
    }
}

// ---------------------------------------------------------------------------------------------------------------------
// ----- Data Rate ------------------------
// ---------------------------------------------------------------------------------------------------------------------
impl DataRate {
    pub const fn from_byte(byte: u8) -> DataRate {
        const SPS2_5: u8 = DataRate::Sps2_5 as u8;
        const SPS5: u8 = DataRate::Sps5 as u8;
        const SPS10: u8 = DataRate::Sps10 as u8;
        const SPS15: u8 = DataRate::Sps15 as u8;
        const SPS25: u8 = DataRate::Sps25 as u8;
        const SPS30: u8 = DataRate::Sps30 as u8;
        const SPS50: u8 = DataRate::Sps50 as u8;
        const SPS60: u8 = DataRate::Sps60 as u8;
        const SPS100: u8 = DataRate::Sps100 as u8;
        const SPS500: u8 = DataRate::Sps500 as u8;
        const SPS1000: u8 = DataRate::Sps1000 as u8;
        const SPS2000: u8 = DataRate::Sps2000 as u8;
        const SPS3750: u8 = DataRate::Sps3750 as u8;
        const SPS7500: u8 = DataRate::Sps7500 as u8;
        const SPS15000: u8 = DataRate::Sps15000 as u8;
        const SPS30000: u8 = DataRate::Sps30000 as u8;

        match byte {
            SPS2_5 => DataRate::Sps2_5,
            SPS5 => DataRate::Sps5,
            SPS10 => DataRate::Sps10,
            SPS15 => DataRate::Sps15,
            SPS25 => DataRate::Sps25,
            SPS30 => DataRate::Sps30,
            SPS50 => DataRate::Sps50,
            SPS60 => DataRate::Sps60,
            SPS100 => DataRate::Sps100,
            SPS500 => DataRate::Sps500,
            SPS1000 => DataRate::Sps1000,
            SPS2000 => DataRate::Sps2000,
            SPS3750 => DataRate::Sps3750,
            SPS7500 => DataRate::Sps7500,
            SPS15000 => DataRate::Sps15000,
            SPS30000 => DataRate::Sps30000,
            _ => panic!("Invalid ADS1256 data rate."),
        }
    }
}

// ---------------------------------------------------------------------------------------------------------------------
// ----- GPIO ------------------------
// ---------------------------------------------------------------------------------------------------------------------
#[derive(Copy, Clone, Eq, PartialEq)]
#[repr(transparent)]
pub struct DigitalIo(pub u8);

#[derive(Copy, Clone, Eq, PartialEq)]
#[repr(transparent)]
pub struct DigitalIoState(pub u8);

impl DigitalIoState {
    #[inline]
    pub const fn new(io0: DState, io1: DState, io2: DState, io3: DState) -> DigitalIoState {
        DigitalIoState(io0 as u8 | ((io1 as u8) << 1) | ((io2 as u8) << 2) | ((io3 as u8) << 3))
    }

    #[inline(always)]
    pub const fn default() -> DigitalIoState {
        DigitalIoState::new(DState::Low, DState::Low, DState::Low, DState::Low)
    }
}

impl Into<u8> for DigitalIoState {
    #[inline(always)]
    fn into(self) -> u8 {
        self.0
    }
}

#[derive(Copy, Clone, Eq, PartialEq)]
#[repr(transparent)]
pub struct DigitalIoDirection(pub u8);

impl DigitalIoDirection {
    #[inline]
    pub const fn new(
        io0: DioDirection,
        io1: DioDirection,
        io2: DioDirection,
        io3: DioDirection,
    ) -> DigitalIoDirection {
        DigitalIoDirection(
            ((io0 as u8) << 4) | ((io1 as u8) << 5) | ((io2 as u8) << 6) | ((io3 as u8) << 7),
        )
    }

    #[inline(always)]
    pub const fn default() -> DigitalIoDirection {
        DigitalIoDirection::new(
            DioDirection::Output,
            DioDirection::Input,
            DioDirection::Input,
            DioDirection::Input,
        )
    }
}

impl Into<u8> for DigitalIoDirection {
    #[inline(always)]
    fn into(self) -> u8 {
        self.0
    }
}

impl DigitalIo {
    #[inline(always)]
    pub const fn setting(io_state: DigitalIoState, io_direction: DigitalIoDirection) -> DigitalIo {
        DigitalIo(io_state.0 | io_direction.0)
    }

    #[inline(always)]
    pub const fn with_io_state(self, state: DigitalIoState) -> DigitalIo {
        let zeroed_setting = self.0 & !DState::ALL_MASK;
        DigitalIo(zeroed_setting | state.0)
    }

    #[inline(always)]
    pub const fn with_io_direction(self, direction: DigitalIoDirection) -> DigitalIo {
        let zeroed_setting = self.0 & !DioDirection::ALL_MASK;
        DigitalIo(zeroed_setting | direction.0)
    }

    #[inline(always)]
    pub const fn with_io0_state(self, state: DState) -> DigitalIo {
        let zeroed_setting = self.0 & !DState::IO0_MASK;
        DigitalIo(zeroed_setting | state as u8)
    }

    #[inline(always)]
    pub const fn with_io1_state(self, state: DState) -> DigitalIo {
        let zeroed_setting = self.0 & !DState::IO1_MASK;
        DigitalIo(zeroed_setting | ((state as u8) << 1))
    }

    #[inline(always)]
    pub const fn with_io2_state(self, state: DState) -> DigitalIo {
        let zeroed_setting = self.0 & !DState::IO2_MASK;
        DigitalIo(zeroed_setting | ((state as u8) << 2))
    }

    #[inline(always)]
    pub const fn with_io3_state(self, state: DState) -> DigitalIo {
        let zeroed_setting = self.0 & !DState::IO3_MASK;
        DigitalIo(zeroed_setting | ((state as u8) << 3))
    }

    #[inline(always)]
    pub const fn io0_state(self) -> DState {
        match self.0 & DState::IO0_MASK {
            DState::IO0_MASK => DState::High,
            _ => DState::Low,
        }
    }

    #[inline(always)]
    pub const fn io1_state(self) -> DState {
        match self.0 & DState::IO1_MASK {
            DState::IO1_MASK => DState::High,
            _ => DState::Low,
        }
    }

    #[inline(always)]
    pub const fn io2_state(self) -> DState {
        match self.0 & DState::IO2_MASK {
            DState::IO2_MASK => DState::High,
            _ => DState::Low,
        }
    }

    #[inline(always)]
    pub const fn io3_state(self) -> DState {
        match self.0 & DState::IO3_MASK {
            DState::IO3_MASK => DState::High,
            _ => DState::Low,
        }
    }

    #[inline(always)]
    pub const fn io0_direction(self) -> DioDirection {
        match self.0 & DioDirection::IO0_MASK {
            DioDirection::IO0_MASK => DioDirection::Input,
            _ => DioDirection::Output,
        }
    }

    #[inline(always)]
    pub const fn io1_direction(self) -> DioDirection {
        match self.0 & DioDirection::IO1_MASK {
            DioDirection::IO1_MASK => DioDirection::Input,
            _ => DioDirection::Output,
        }
    }

    #[inline(always)]
    pub const fn io2_direction(self) -> DioDirection {
        match self.0 & DioDirection::IO2_MASK {
            DioDirection::IO2_MASK => DioDirection::Input,
            _ => DioDirection::Output,
        }
    }

    #[inline(always)]
    pub const fn io3_direction(self) -> DioDirection {
        match self.0 & DioDirection::IO3_MASK {
            DioDirection::IO3_MASK => DioDirection::Input,
            _ => DioDirection::Output,
        }
    }
}

impl Into<u8> for DigitalIo {
    #[inline(always)]
    fn into(self) -> u8 {
        self.0
    }
}

#[cfg(feature = "defmt")]
impl defmt::Format for DigitalIo {
    fn format(&self, fmt: defmt::Formatter) {
        defmt::write!(
            fmt,
            "DigitalIo(io0: {}, {}, io1: {}, {}, io2: {}, {}, io3: {}, {})",
            self.io0_direction(),
            self.io0_state(),
            self.io1_direction(),
            self.io1_state(),
            self.io2_direction(),
            self.io2_state(),
            self.io3_direction(),
            self.io3_state()
        )
    }
}
