#![no_std]
#![no_main]

use cortex_m::prelude::{_embedded_hal_blocking_delay_DelayMs, _embedded_hal_blocking_delay_DelayUs};
use {defmt_rtt as _, panic_probe as _};

use {embassy_executor as executor, embassy_stm32 as stm32};

use ads1256::{
    AdControl, Ads1256, AutoCal, BitOrder, Buffer, ClockOut, Config, DState, DataRate, DigitalIo,
    DigitalIoDirection, DigitalIoState, DioDirection, Gain, Multiplexer, MuxInput, OutputPin, Sdcs,
    SpiBus, Status, Wait, BlockingDelay
};
use embassy_time::{Delay, Timer};
use executor::Spawner;
use physical::quantity::Quantity;
use stm32::dma::NoDma;
use stm32::exti::ExtiInput;
use stm32::gpio::{Input, Level, Output, Pull, Speed};
use stm32::spi::Spi;
use stm32::time::Hertz;
use stm32::{pac, spi};

use defmt::{debug, error, info, trace, unwrap};

const AUTOCAL_CONF: Config = Config {
    status: Status::setting(Buffer::Enabled, AutoCal::Enabled, BitOrder::MostSigFirst),
    multiplexer: Multiplexer::setting(MuxInput::AIn0, MuxInput::Common),
    ad_control: AdControl::setting(Gain::X2, Sdcs::Off, ClockOut::Off),
    data_rate: DataRate::Sps2_5,
    digital_io: DigitalIo::setting(DigitalIoState::default(), DigitalIoDirection::default()),
};

const ADS1256_DELAY: ads1256::DefaultDelay<Delay> = ads1256::DefaultDelay::new(Delay);

#[embassy_executor::main]
async fn main(spawner: Spawner) {
    unsafe {
        pac::FLASH.acr().modify(|v| {
            v.set_prften(true);
            v.set_icen(true);
            v.set_dcen(true);
        });
    }

    let p = embassy_stm32::init(Default::default());

    let mut spi_conf = spi::Config::default();
    spi_conf.mode = spi::MODE_1;
    spi_conf.bit_order = spi::BitOrder::MsbFirst;
    spi_conf.frequency = Hertz(ads1256::defaults::SPI_CLK_HZ);

    let ads1256_data_ready = ExtiInput::new(Input::new(p.PA3, Pull::Up), p.EXTI3);
    let select_ads1256 = Output::new(p.PA1, Level::High, Speed::VeryHigh);

    let mut spi = Spi::new(
        p.SPI1,
        p.PA5,
        p.PA7,
        p.PA6,
        NoDma,
        NoDma,
        spi_conf,
    );

    let mut ads_1256 = Ads1256::new(ADS1256_DELAY, select_ads1256, ads1256_data_ready);
    // single_conversion(&mut spi, &mut ads_1256).await;
    // ads_1256.delayer.t11_1_delay();
    // read_continuous(&mut spi, &mut ads_1256).await;
    cycle_multiplexer(&mut spi, &mut ads_1256).await;
}

async fn single_conversion<DelayerT: ads1256::BlockingDelay, SST: OutputPin, DrdyT: Wait>(
    spi: &mut impl SpiBus,
    ads_1256: &mut Ads1256<DelayerT, SST, DrdyT>,
) {
    ads_1256.write_config(spi, AUTOCAL_CONF).unwrap();
    ads_1256.delayer.t11_1_delay();
    ads_1256.conversion_init(spi).unwrap();
    let data = ads_1256.cmd_read_data(spi).await.unwrap();
    info!("data: {}, volts: {}", data, data.to_voltage(AUTOCAL_CONF.ad_control.gain()).fmt(Some(5)));
}

async fn read_continuous<DelayerT: ads1256::BlockingDelay, SST: OutputPin, DrdyT: Wait>(
    spi: &mut impl SpiBus,
    ads_1256: &mut Ads1256<DelayerT, SST, DrdyT>,
) {
    ads_1256.write_config(spi, AUTOCAL_CONF).unwrap();
    ads_1256.delayer.t11_1_delay();
    ads_1256.start_rdatac(spi).await.unwrap();
    loop {
        let data = ads_1256.read_data(spi).await.unwrap();
        info!("data: {}, volts: {}", data, data.to_voltage(AUTOCAL_CONF.ad_control.gain()).fmt(Some(5)));
    }
}

async fn cycle_multiplexer<DelayerT: ads1256::BlockingDelay, SST: OutputPin, DrdyT: Wait>(
    spi: &mut impl SpiBus,
    ads_1256: &mut Ads1256<DelayerT, SST, DrdyT>,
) {
    const INPUT_1: Multiplexer = Multiplexer::setting(MuxInput::AIn1, MuxInput::Common);
    const INPUT_2: Multiplexer = Multiplexer::setting(MuxInput::AIn2, MuxInput::Common);
    const INPUT_3: Multiplexer = Multiplexer::setting(MuxInput::AIn3, MuxInput::Common);

    let ad_control = AUTOCAL_CONF.ad_control;
    let status = AUTOCAL_CONF.status;

    ads_1256.write_config(spi, AUTOCAL_CONF).unwrap();
    ads_1256.delayer.t11_1_delay();
    loop {
        let ad_control = ad_control.with_gain(Gain::X4);
        let data = ads_1256
            .autocal_convert(spi, INPUT_1, None, Some(ad_control.with_gain(Gain::X4)), None, false)
            .await
            .unwrap();
        info!("Input 1: data: {}, volts: {}", data, data.to_voltage(ad_control.gain()).fmt(Some(5)));
        let ad_control = ad_control.with_gain(Gain::X8);
        let data = ads_1256
            .autocal_convert(spi, INPUT_2, None, Some(ad_control.with_gain(Gain::X8)), None, false)
            .await
            .unwrap();
        info!("Input 2: data: {}, volts: {}", data, data.to_voltage(ad_control.gain()).fmt(Some(5)));
        let ad_control = ad_control.with_gain(Gain::X16);
        let data = ads_1256
            .autocal_convert(spi, INPUT_3, None, Some(ad_control.with_gain(Gain::X16)), None, false)
            .await
            .unwrap();
        info!("Input 3: data: {}, volts: {}", data, data.to_voltage(ad_control.gain()).fmt(Some(5)));
    }

}
