#![no_std]
#![no_main]

use cortex_m::prelude::_embedded_hal_blocking_delay_DelayUs;
use {defmt_rtt as _, panic_probe as _};

use {embassy_executor as executor, embassy_stm32 as stm32};

use ads1256::{
    AdControl, Ads1256, AutoCal, BitOrder, Buffer, ClockOut, Config, DState, DataRate, DigitalIo,
    DigitalIoDirection, DigitalIoState, DioDirection, Gain, Multiplexer, MuxInput, OutputPin, Sdcs,
    SpiBus, Status, Wait,
};
use embassy_time::Delay;
use executor::Spawner;
use stm32::dma::NoDma;
use stm32::exti::ExtiInput;
use stm32::gpio::{Input, Level, Output, Pull, Speed};
use stm32::spi::Spi;
use stm32::time::Hertz;
use stm32::{pac, spi};

use defmt::{debug, error, info, trace, unwrap};

const ADS1256_DELAY: ads1256::DefaultDelay<Delay> = ads1256::DefaultDelay::new(Delay);

#[embassy_executor::main]
async fn main(spawner: Spawner) {
    unsafe {
        pac::FLASH.acr().modify(|v| {
            v.set_prften(true);
            v.set_icen(true);
            v.set_dcen(true);
        });
    }

    let p = embassy_stm32::init(Default::default());

    let mut spi_conf = spi::Config::default();
    spi_conf.mode = spi::MODE_1;
    spi_conf.bit_order = spi::BitOrder::MsbFirst;
    spi_conf.frequency = Hertz(ads1256::defaults::SPI_CLK_HZ);

    let ads1256_data_ready = ExtiInput::new(Input::new(p.PA3, Pull::Up), p.EXTI3);
    let select_ads1256 = Output::new(p.PA1, Level::High, Speed::VeryHigh);

    let mut spi = Spi::new(
        p.SPI1,
        p.PA5,
        p.PA7,
        p.PA6,
        NoDma,
        NoDma,
        spi_conf,
    );

    let mut ads_1256 = Ads1256::new(ADS1256_DELAY, select_ads1256, ads1256_data_ready);
    // status(&mut spi, &mut ads_1256);
    // multiplexer(&mut spi, &mut ads_1256);
    // ad_control(&mut spi, &mut ads_1256);
    // data_rate(&mut spi, &mut ads_1256);
    // gpio(&mut spi, &mut ads_1256);
     config(&mut spi, &mut ads_1256);
}

fn status<DelayerT: ads1256::BlockingDelay, SST: OutputPin, DrdyT: Wait>(
    spi: &mut impl SpiBus,
    ads_1256: &mut Ads1256<DelayerT, SST, DrdyT>,
) {
    info!("Status register:");
    const STATUS_SETTING: Status =
        Status::setting(Buffer::Disabled, AutoCal::Disabled, BitOrder::MostSigFirst);
    ads_1256.write_status(spi, STATUS_SETTING).unwrap();
    let status = ads_1256.read_status(spi).unwrap();
    info!("ADS1256 starting status: {}", status);
    let new_status_setting = status.with_buffer(Buffer::Enabled);
    ads_1256.write_status(spi, new_status_setting).unwrap();
    let status = ads_1256.read_status(spi).unwrap();
    info!("ADS1256 new status: {}", status);
}

fn multiplexer<DelayerT: ads1256::BlockingDelay, SST: OutputPin, DrdyT: Wait>(
    spi: &mut impl SpiBus,
    ads_1256: &mut Ads1256<DelayerT, SST, DrdyT>,
) {
    info!("Multiplexer register:");
    const MULTIPLEXER_SETTING: Multiplexer = Multiplexer::setting(MuxInput::AIn0, MuxInput::Common);
    ads_1256
        .write_multiplexer(spi, MULTIPLEXER_SETTING)
        .unwrap();
    let multiplexer = ads_1256.read_multiplexer(spi).unwrap();
    info!("ADS1256 starting multiplexer: {}", multiplexer);
    let new_multiplexer_setting = multiplexer.with_positive(MuxInput::AIn1);
    ads_1256
        .write_multiplexer(spi, new_multiplexer_setting)
        .unwrap();
    let multiplexer = ads_1256.read_multiplexer(spi).unwrap();
    info!("ADS1256 new ad_control: {}", multiplexer);
}

fn ad_control<DelayerT: ads1256::BlockingDelay, SST: OutputPin, DrdyT: Wait>(
    spi: &mut impl SpiBus,
    ads_1256: &mut Ads1256<DelayerT, SST, DrdyT>,
) {
    info!("AD control register:");
    const AD_CONTROL_SETTING: AdControl = AdControl::setting(Gain::X64, Sdcs::Off, ClockOut::Off);
    ads_1256.write_ad_control(spi, AD_CONTROL_SETTING).unwrap();
    let ad_control = ads_1256.read_ad_control(spi).unwrap();
    info!("ADS1256 starting ad_control: {}", ad_control);
    let new_ad_control_setting = ad_control.with_gain(Gain::X1);
    ads_1256
        .write_ad_control(spi, new_ad_control_setting)
        .unwrap();
    let ad_control = ads_1256.read_ad_control(spi).unwrap();
    info!("ADS1256 new ad_control: {}", ad_control);
}

fn data_rate<DelayerT: ads1256::BlockingDelay, SST: OutputPin, DrdyT: Wait>(
    spi: &mut impl SpiBus,
    ads_1256: &mut Ads1256<DelayerT, SST, DrdyT>,
) {
    info!("Data rate register:");
    const DATA_RATE: DataRate = DataRate::Sps2_5;
    ads_1256.write_data_rate(spi, DATA_RATE).unwrap();
    let data_rate = ads_1256.read_data_rate(spi).unwrap();
    info!("ADS1256 starting data rate: {}", data_rate);
    ads_1256.write_data_rate(spi, DataRate::Sps60).unwrap();
    let data_rate = ads_1256.read_data_rate(spi).unwrap();
    info!("ADS1256 new data rate: {}", data_rate);
}

fn gpio<DelayerT: ads1256::BlockingDelay, SST: OutputPin, DrdyT: Wait>(
    spi: &mut impl SpiBus,
    ads_1256: &mut Ads1256<DelayerT, SST, DrdyT>,
) {
    info!("GPIO register:");
    const GPIO: DigitalIo = DigitalIo::setting(
        DigitalIoState::new(DState::Low, DState::Low, DState::Low, DState::Low),
        DigitalIoDirection::new(
            DioDirection::Output,
            DioDirection::Input,
            DioDirection::Input,
            DioDirection::Output,
        ),
    );
    ads_1256.write_gpio(spi, GPIO).unwrap();
    let gpio = ads_1256.read_gpio(spi).unwrap();
    info!("ADS1256 starting gpio: {}", gpio);
    let new_gpio_setting = gpio.with_io3_state(DState::High);
    ads_1256.write_gpio(spi, new_gpio_setting).unwrap();
    let gpio = ads_1256.read_gpio(spi).unwrap();
    info!("ADS1256 new gpio: {}", gpio);
}

fn config<DelayerT: ads1256::BlockingDelay, SST: OutputPin, DrdyT: Wait>(
    spi: &mut impl SpiBus,
    ads_1256: &mut Ads1256<DelayerT, SST, DrdyT>,
) {
    info!("Config:");
    let config = ads_1256.read_config(spi).unwrap();
    info!("ADS1256 starting config: {}", config);
    const CONFIG: Config = Config {
        status: Status::setting(Buffer::Enabled, AutoCal::Enabled, BitOrder::MostSigFirst),
        multiplexer: Multiplexer::setting(MuxInput::AIn4, MuxInput::AIn5),
        ad_control: AdControl::setting(Gain::X16, Sdcs::Off, ClockOut::Off),
        data_rate: DataRate::Sps10,
        digital_io: DigitalIo::setting(
            DigitalIoState::new(DState::Low, DState::Low, DState::Low, DState::Low),
            DigitalIoDirection::new(
                DioDirection::Output,
                DioDirection::Input,
                DioDirection::Input,
                DioDirection::Input,
            ),
        ),
    };
    ads_1256.write_config(spi, CONFIG).unwrap();
    let config = ads_1256.read_config(spi).unwrap();
    info!("ADS1256 new config: {:#?}", config);
}
