use generate_quantity::quantity_type;
use crate::quantity::{DECI, Quantity, Value};

const KELVIN_CELSIUS_OFFSET: f64 = 273.15;

//----- Kelvins ----------------------------------
quantity_type! {Kelvins, "K"}

impl <V: Value> Kelvins<V> {
    #[inline]
    pub fn to_celsius(self) -> Celsius<V> {
        //TODO: Make const
        let offset = V::from_f64(KELVIN_CELSIUS_OFFSET).unwrap();
        Celsius(self.0 - offset)
    }

    #[inline]
    pub fn to_deci_kelvins(self) -> DeciKelvins<V> {
        let multiplier = V::from_u8(DECI).unwrap();
        DeciKelvins(self.0 * multiplier)
    }
}

//----- Decikelvins ----------------------------------
quantity_type! {DeciKelvins, "dK"}

impl<V: Value> DeciKelvins<V> {
    #[inline]
    pub fn to_kelvins(self) -> Kelvins<V> {
        let divisor = V::from_u8(DECI).unwrap();
        Kelvins(self.0 / divisor)
    }
}

//----- Degrees Celsius ----------------------------------
quantity_type! {Celsius, "℃"}

impl <V: Value> Celsius<V> {
    #[inline]
    pub fn to_kelvins(self) -> Kelvins<V> {
        //TODO: Make const
        let offset = V::from_f64(KELVIN_CELSIUS_OFFSET).unwrap();
        Kelvins(self.0 + offset)
    }

    #[inline]
    pub fn to_deci_celsius(self) -> DeciCelsius<V> {
        let multiplier = V::from_u8(DECI).unwrap();
        DeciCelsius(self.0 * multiplier)
    }
}

//----- Deci Degrees Celsius ----------------------------------
quantity_type! {DeciCelsius, "d℃"}

impl<V: Value> DeciCelsius<V> {
    #[inline]
    pub fn to_celsius(self) -> Celsius<V> {
        let divisor = V::from_u8(DECI).unwrap();
        Celsius(self.0 / divisor)
    }
}

// ---------------------------------------------------------------------------------------------------------------------
// ----- Tests ------------------------
// ---------------------------------------------------------------------------------------------------------------------
#[cfg(test)]
mod tests {
    use float_cmp::assert_approx_eq;
    use super::*;

    #[test]
    fn convert_f32() {
        let kelvins: Kelvins<f32> =  298.15.kelvins();
        let deci_kelvins: DeciKelvins<f32> = 2_981.5.deci_kelvins();
        let celsius: Celsius<f32> = 25.0.celsius();
        let deci_celsius: DeciCelsius<f32> = 250.0.deci_celsius();

        assert_approx_eq!(f32, kelvins.to_celsius().0, celsius.0);
        assert_approx_eq!(f32, celsius.to_kelvins().0, kelvins.0);

        assert_approx_eq!(f32, deci_kelvins.to_kelvins().0, kelvins.0);
        assert_approx_eq!(f32, kelvins.to_deci_kelvins().0, deci_kelvins.0);

        assert_approx_eq!(f32, deci_celsius.to_celsius().0, celsius.0);
        assert_approx_eq!(f32, celsius.to_deci_celsius().0, deci_celsius.0);
    }

    #[test]
    fn convert_u16() {
        let kelvins: Kelvins<u16> = 298.kelvins();
        let celsius: Celsius<u16> = 25.celsius();

        assert_eq!(celsius.0, kelvins.to_celsius().0);
    }
}
